/@
    CAN 4 Linux, a CAN bus driver for Linux.
@/
module "Can" : 

/*include Lib.ddl*/

driver "can" [ major=63 ] {

header "config" /{
#if 1
#define FAST_XMIT 
#endif
}/


include 82c200.ddl
include intel.ddl
include can_proto.ddl
include util.ddl
include chipset.ddl
/*include pager.ddl*/
include Timer.ddl

	sysctl {

	   /*
            * this are the configurable parameters exported to /proc/sys/Can
            *
            *
            */
	   string version [access=0444]     = "0.8 alpha"
           string IOModel [access=0644]     = "pppp "
           string TxSpeed [access=0644]     = "ffff "
           string VendOpt [access=0644]     = "nnnn "
	   int[4] IRQ     [access=0644]     = 0
	   int[4] Base    [access=0644]     = 0x320
	   int[4] Baud    [access=0644]     = 125
	   int[4] AccMask [access=0644]     = 0xffff
	   int[4] Timeout [access=0644]     = 100
           int[4] Outc    [access=0644]     = 0

	   int[4] TxErr   [access=0444]     = 0
           int[4] RxErr   [access=0444]     = 0
	   int[4] Overrun [access=0444]     = 0

	   int[4] Chipset [ handler=ChipsetHandler ]     = 0

           handler ChipsetHandler /{
		int *i, vleft, first=1, len, left, neg, val;
		#define TMPBUFLEN 20
		char buf[TMPBUFLEN], *p;
	
		if (!table->data || !table->maxlen || !*lenp ||
		    (filp->f_pos && !write)) {
			*lenp = 0;
			return 0;
		}
	
		i = (int *) table->data;
		vleft = table->maxlen / sizeof(int);
		left = *lenp;
	
		for (; left && vleft--; i++, first=0) {
			if (write) {
				while (left && isspace(get_user((char *) buffer)))
					left--, ((char *) buffer)++;
				if (!left)
					break;
			neg = 0;
			len = left;
			if (len > TMPBUFLEN-1)
				len = TMPBUFLEN-1;
			memcpy_fromfs(buf, buffer, len);
			buf[len] = 0;
			p = buf;

			if (buf[0] == '8' && !strncmp(buf,"82c200",6) ) {
				val = CHIPSET_82c200;
				p+=6;
				/*printk("chipset 82c200\n");*/
                        }

			if (buf[0] == 'i' && !strncmp(buf,"i82527",6) ) {
				val = CHIPSET_i82527;
				p+=6;
				/*printk("chipset i82527\n");*/
			}
			if (buf[0] == 'e' && !strncmp(buf,"ems-cpc",7) ) {
				val = CHIPSET_emscpc;
				p+=7;
				/*printk("chipset EMS-cpc\n");*/
			}
			if (buf[0] == 'p' && !strncmp(buf,"pccan",5) ) {
				val = CHIPSET_pccan;
				p+=5;
				/*printk("chipset EMS-cpc\n");*/
			}
			if (buf[0] == 'e' && !strncmp(buf,"emu",3) ) {
				val = CHIPSET_emu;
				p+=3;
				/*printk("chipset i82527\n");*/
			}
			if (buf[0] == 'n' && !strncmp(buf,"none",4) ) {
				val = CHIPSET_none;
				p+=4;
				/*printk("no chipset configured!\n");*/
			}

			len = p-buf;
			if ((len < left) && *p && !isspace(*p))
				break;
			buffer += len;
			left -= len;
			*i = val;

		} else {
			p = buf;
			if (!first)
				*p++ = '\t';
			switch( *i ) {
				case CHIPSET_none:
					sprintf(p, "none"); break;
				case CHIPSET_82c200:
					sprintf(p, "82c200"); break;
				case CHIPSET_i82527:
					sprintf(p, "i82527"); break;
				case CHIPSET_emscpc:
					sprintf(p, "ems-cpc"); break;
				case CHIPSET_emu:
					sprintf(p, "emu"); break;
				case CHIPSET_pccan:
					sprintf(p, "pccan"); break;
				default:
					sprintf(p, "unknown"); break;

			}
			len = strlen(buf);
			if (len > left)
				len = left;
			memcpy_tofs(buffer, buf, len);
			left -= len;
			buffer += len;
		}
	        }

		if (!write && !first && left) {
			put_user('\n', (char *) buffer);
			left--, buffer++;
		}
		if (write) {
			p = (char *) buffer;
		while (left && isspace(get_user(p++)))
			left--;
		}
		if (write && first)
			return -EINVAL;
		*lenp -= left;
		filp->f_pos += *lenp;
		DBGout();return 0;
	   }/

	}

	core {  
		init /{ 
			Base[0] = 0x320;
			Base[1] = 0x340;
			Base[2] = 0x280;
			Base[3] = 0x380;
                        

			printk("CAN Driver %s (c) %<date> \n",version);
			printk(" C.Schroeter (clausi@chemie.fu-berlin.de)\n"); 
			printk(" H.J. Oertel (oe@port.de)\n");
			printk(" H.D. Stich  ()\n");
		}/

		cleanup /{
			printk("CAN: removed successfully \n");
		}/
	}

	open /{
		extern void Can_Interrupt(int, void *, struct pt_regs *);
		extern can_atomic_func_t p82c200_ops;	
		extern can_atomic_func_t emscpc_ops;	
		extern can_atomic_func_t pccan_ops;	
		extern can_atomic_func_t i527_ops;	
                int lasterr;	

		unsigned int minor = MINOR(inode->i_rdev);

		if( minor > 4 ) {
		   printk("CAN: Illegal minor number %d\n", minor );
		   MOD_DEC_USE_COUNT;
		   DBGout();
		   return -EINVAL;
		}

		/* choose chipset type */

                switch(Chipset[minor]) {
                 case CHIPSET_none:
			printk("Can[%d]: No Chipset: Configure Chipset first!\n",minor);
                        MOD_DEC_USE_COUNT;
		        DBGout();return -EINVAL;
                        break;			
		 case CHIPSET_82c200:
			DBGprint(DBG_BRANCH,("Can[%d]: Chipset: 82c200",minor));
			atomic_ops[minor] = &p82c200_ops;
			break;
                 case CHIPSET_emscpc:
			DBGprint(DBG_BRANCH,("Can[%d]: Chipset: ems-cpc",minor));
			atomic_ops[minor] = &emscpc_ops;
			break;
                 case CHIPSET_pccan:
			DBGprint(DBG_BRANCH,("Can[%d]: Chipset: kvaser pccan",minor));
			atomic_ops[minor] = &pccan_ops;
			break;
                 case CHIPSET_i82527:
			DBGprint(DBG_BRANCH,("Can[%d]: Chipset: generic i82527",minor));
			atomic_ops[minor] = &i527_ops;
			break;
		 default:
			printk("Can[%d]: Unsupported Chipset type %d !\n",minor, Chipset[minor]);
                        MOD_DEC_USE_COUNT;
		        DBGout();return -EINVAL;			
		}

                if( strlen(IOModel) < 4 ) {
			printk("Can[%d]: Misconfigured IO Model ",minor);
                        MOD_DEC_USE_COUNT;
		        DBGout();return -EINVAL;			
		} else {
			
			Set_IO_Model(minor, IOModel[minor] );
		}
                if( (lasterr=CAN_VendorInit(minor)) < 0 ){
		   	MOD_DEC_USE_COUNT;
		   	DBGout();
		   	return lasterr;
		}

		Can_FifoInit(minor);
		Can_FilterInit(minor);

		if( CAN_ChipReset(minor) < 0 ) {
		   MOD_DEC_USE_COUNT;
		   DBGout();
		   return -EINVAL;
		}
		CAN_StartChip(minor);
	}/

	close /{
		unsigned int minor = MINOR(inode->i_rdev);
                extern int CAN_Open;
		CAN_StopChip(minor);
                
                if( Chipset[minor] == CHIPSET_pccan ) { 
                  release_region(Base[0],CAN_RANGE );
                }
                else
		  release_region(Base[minor],CAN_RANGE );
                if( Chipset[minor] == CHIPSET_emscpc || Chipset[minor]== CHIPSET_pccan ) CAN_Open--;
		Can_FilterCleanup(minor);

		if( IRQ[minor] > 0 ) 
                      if( Chipset[minor] == CHIPSET_emscpc || Chipset[minor]== CHIPSET_pccan ) {
                       
                        if ( CAN_Open <= 0 )
                           Can_FreeIrq(minor, IRQ[minor] );
                      } 
                       else  
                      Can_FreeIrq(minor, IRQ[minor] );

		printk("CAN module has been closed\n");
	}/

	read /{
                unsigned int minor = MINOR(inode->i_rdev);
		msg_fifo_t *RxFifo = &Rx_Buf[minor];
		canmsg_t *addr; 
		int written=0;

		addr = buffer;
                
		if( verify_area( VERIFY_WRITE, buffer, count * sizeof(canmsg_t) )) {
		   DBGout();return -EINVAL;
		}
		while( written < count && RxFifo->status == BUF_OK ) {
		   if( RxFifo->tail == RxFifo->head ) {
			RxFifo->status = BUF_EMPTY;
			break;
		   }	

		   memcpy_tofs( (canmsg_t *) &addr[written], 
                                (canmsg_t *) &(RxFifo->data[RxFifo->tail]),
				sizeof(canmsg_t) );
		   written++;
		   RxFifo->tail = ++(RxFifo->tail) % MAX_BUFSIZE;
		}
 		DBGout(); return(written);
	}/
        
	write /{
		unsigned int minor = MINOR(inode->i_rdev);
                msg_fifo_t *TxFifo = &Tx_Buf[minor];
		canmsg_t *addr;
                int written=0;
                canmsg_t tx;

                addr = buffer;
		if( verify_area( VERIFY_READ, (canmsg_t *) addr, count * sizeof(canmsg_t))) { 
			DBGout();return -EINVAL;
		}
                while( written < count ) {

		  /*CANreset(minor, canctl, CAN_TRANSMIT_INT_ENABLE | CAN_OVERRUN_INT_ENABLE );*/
		  if( TxFifo->active && count > 1 ) {		
			memcpy_fromfs( (canmsg_t *) &(TxFifo->data[TxFifo->head]), 
                                                    (canmsg_t *) &addr[written], sizeof(canmsg_t) );
			TxFifo->head = ++(TxFifo->head) % MAX_BUFSIZE;
 			if( TxFifo->head == TxFifo->tail ) {
			  printk("%d TX: FIFO underrun !\n",minor);
			  TxFifo->status = BUF_UNDERRUN;
			  DBGout();return -1;
			}
			/*CANset(minor, canctl, CAN_TRANSMIT_INT_ENABLE | CAN_OVERRUN_INT_ENABLE );*/
		  } else {
                        memcpy_fromfs( (canmsg_t *) &tx, 
                                       (canmsg_t *) &addr[written], sizeof(canmsg_t) );
			if( TxSpeed[minor] == 'f' ) {
                          if( count > 1 ) {
				  CANset(minor, canctl, CAN_TRANSMIT_INT_ENABLE | CAN_OVERRUN_INT_ENABLE );
				  TxFifo->active = 1;
			  }
			  CAN_SendMessage( minor, &tx, 0);
                        } else {
			  CAN_SendMessage( minor, &tx, 1);
                        }
		  }
		  written++;
		}
		/*CANreset(minor, canctl, CAN_TRANSMIT_INT_ENABLE | CAN_OVERRUN_INT_ENABLE );*/
                DBGout();return written;
	}/

        select /{
                unsigned int minor = MINOR(inode->i_rdev);
		msg_fifo_t *RxFifo = &Rx_Buf[minor];
		extern struct wait_queue *CanWait[];

                switch (sel_type) {

                        case SEL_IN:
                             DBGprint(DBG_BRANCH,("sel_in \n"));
                             if( RxFifo->head == RxFifo->tail ) {
				DBGprint(DBG_BRANCH,("fifo empty \n"));
                                select_wait(&CanWait[minor],wait);
                                DBGout();return 0;
			     }
                             break;
                        case SEL_OUT:
                             DBGprint(DBG_BRANCH,("sel_out \n"));
                             /* ready for write ? */
                             select_wait(&CanWait[minor],wait);
                             DBGout();return 0;
                }
		DBGout();return 1;

        }/


	ioctl {
	     /@
		Request Can command:

                   cmd=
                   ---------------------------------------------------------
		   CMD_START  -  start the CAN controller
		   CMD_STOP   -  sends a stop request
		   CMD_RESET  -  Chip Reset request 

	     @/
	     func Command (int cmd) 
	     /{
		unsigned int minor = MINOR(inode->i_rdev);

		DBGprint(DBG_DATA,("cmd=%d",cmd));
		switch (cmd) {
		  case CMD_START:
			CAN_StartChip(minor);
		  case CMD_STOP:
			CAN_StopChip(minor);
		  case CMD_RESET:
			CAN_ChipReset(minor);
		  default:
			DBGout();
			return(-EINVAL);
		}
	     }/

	     /@
		Configure Can Board parameters:

                   target =                                   arg format  
		   ----------------------------------------------------------
                   Global settings

		   CONF_ACCM        set chip acceptance mask  0xffff
		   CONF_TIMING      set baudrate              10,20,40,50,125,...
		   CONF_OMODE       set transceiver mode      0xff


                   COB's settings

		   CONF_FILTER      set software filter       on/of 1,0
		   CONF_FENABLE     set pass id on	      CANid
		   CONF_FDISABLE    set pass id off           CANid

              @/
	     func Config(int target, unsigned long val ) 
             /{
	        unsigned int minor = MINOR(inode->i_rdev);	

		switch(target) {
		  case CONF_ACCM:
		       AccMask[minor] = val;
		       CAN_SetMask(minor, (int) val );		
		       break;
		  case CONF_TIMING:
		       Baud[minor] = val;
		       CAN_SetTiming(minor,(int) val);   
		       break;                    
		  case CONF_OMODE:
		       CAN_SetOMode( minor, (int) val);
		       break;			
		  case CONF_FILTER:
		       Can_FilterOnOff( minor, (int) val );
		       break;
                  case CONF_FENABLE:
		       Can_FilterMessage( minor, (int) val, 1);
		       break;
		  case CONF_FDISABLE:
		       Can_FilterMessage( minor, (int) val, 0);
		       break;



		  case CONF_XTDACCM:
			if( Chipset[minor] == CHIPSET_i82527 ) {
				Can_i527_SetXTDMask(minor,val);
			} else {
				printk("Can[%d]: no capability for setting i527 XTD mask ");
				DBGout(); return -EINVAL;
			}	

		  default:
			DBGout();
			return(-EINVAL);
		}
	     }/

            func Send ( canmsg_t *Tx ) /{
		unsigned int minor = MINOR(inode->i_rdev);	
		canmsg_t tx;
                int i;

		if( verify_area(VERIFY_READ,Tx,sizeof(canmsg_t)) ) {
			return -EINVAL;
		}
		memcpy_fromfs((canmsg_t *) &tx, (canmsg_t *) Tx,sizeof(canmsg_t));
		return CAN_SendMessage(minor,&tx, 1);

	    }/

	    func Receive ( canmsg_t *Rx ) /{
		unsigned int minor = MINOR(inode->i_rdev);
		canmsg_t rx;
		int len;

		len=CAN_GetMessage(minor,&rx);

		if( len > 0 ){
 		   printk("\nrx[ ] got id 0x%x len %d adr00x%lx\n",rx.id,rx.length,Rx);
		   if( verify_area(VERIFY_WRITE,Rx,sizeof(canmsg_t) ) ) {
			return -EINVAL;
                   }
		   memcpy_tofs((canmsg_t *) Rx,(canmsg_t *) &rx ,sizeof(canmsg_t));
		   return rx.length;
		} else {
                 /* no message availiable */
		 return 0;
		}

	    }/

	    func ConfigureRTR ( unsigned message, canmsg_t *Tx ) /{
		unsigned int minor = MINOR(inode->i_rdev);
		canmsg_t msg;
		DBGin("ioctl:ConfigureRTR");
		if( Tx != NULL ){
		if( verify_area(VERIFY_READ,Tx,sizeof(canmsg_t)) )
			{DBGout();return -EINVAL;}
		memcpy_fromfs( (canmsg_t *) &msg , (canmsg_t *) Tx, sizeof(canmsg_t) );

		Can_ConfigRTR( minor, message, &msg );
		} else {
		Can_UnConfigRTR( minor, message );
		}
		DBGout();
	    }/

/*
	    func Status() /{
		unsigned int minor = MINOR(inode->i_rdev);
		int status;
		DBGin("Status");
		status = CANin(minor,canstat);
		DBGprint(DBG_DATA,("stat=0x%x",status));
		DBGout(); return status;	
	    }/

	    func Chipstatus ( ) /{
		unsigned int minor = MINOR(inode->i_rdev);
		int i;
		printk("---- 0x%x --------------------------------------- \n",Base[minor]);		
		printk("CNTL: 0x%x \n", CANin(minor,canctl  ));
		printk("STAT: 0x%x \n", CANin(minor,canstat ));
		printk("TXD1: 0x%x \n", CANin(minor,cantxdes1  ));
		printk("TXD2: 0x%x \n", CANin(minor,cantxdes2  ));
		printk("TXd : ");
		for(i=0;i<8;i++)
		   printk("[0x%0x]", 
				CANin(minor,cantxdata[i] ));
		
		printk("\nRXD1: 0x%x \n", CANin(minor,canrxdes1  ));
		printk("RXD2: 0x%x \n", CANin(minor,canrxdes2  ));
		printk("RXd : ");
		for(i=0;i<8;i++)
		   printk("[0x%0x]", 
				CANin(minor,canrxdata[i] ));
		printk("\n-----------------------------------------------\n");

	    }/
*/
        }
}

