/*
 *  Support file for the Philips 82c200 basic can controller
 *
 *  - currently tested for PORT At-mini-can boards
 *  - support for EMS CPC card
 *
 */

header "82c200conf" /{
  /*
   * default configuration
   *
   */

/*---------- Default Outc value 
 * this depends on the transceiver configuration
 * 
 * the port board uses optocoupler configuration as denoted in the philips
 * application notes, so the OUTC value is 0xfa
 *
 */

#define CAN_OUTC_VAL            0xfa
#define CPC_OUTC_VAL            0xda
#define PCCAN_OUTC_VAL          (CAN_OCTP1 | CAN_OCTN1 | CAN_OCTP0 | CAN_OCTN0 | CAN_OCMODE1)

#define USE_FILTER              1


}/

header "portio" /{

static inline void wait_cycle( void )
{
  __asm__ volatile ("outb %al,$0x80");
}

#	define CANout(bd,adr,v) (io_funcs[bd].OutB)(v, (int) &((canregs_t *)Base[bd])->adr )
#	define CANin(bd,adr)    (io_funcs[bd].InB) ((int) &((canregs_t *)Base[bd])->adr  )
#	define CANset(bd,adr,m) (io_funcs[bd].OutB)(((io_funcs[bd].InB)((int) &((canregs_t *)Base[bd])->adr)) | m ,(int) &((canregs_t *)Base[bd])->adr )
#	define CANreset(bd,adr,m) (io_funcs[bd].OutB)(((io_funcs[bd].InB)((int) &((canregs_t *)Base[bd])->adr)) & ~m,(int) &((canregs_t *)Base[bd])->adr )
#	define CANtest(bd,adr,m) ((io_funcs[bd].InB)((int) &((canregs_t *)Base[bd])->adr  ) & m )


#if OLD_MODEL
#	define CANout(bd,adr,v) (atomic_ops[bd]->OutB)(v, (int) &((canregs_t *)Base[bd])->adr )
#	define CANin(bd,adr)    (atomic_ops[bd]->InB) ((int) &((canregs_t *)Base[bd])->adr  )
#	define CANset(bd,adr,m) (atomic_ops[bd]->OutB)(((atomic_ops[bd]->InB)((int) &((canregs_t *)Base[bd])->adr)) | m ,(int) &((canregs_t *)Base[bd])->adr )
#	define CANreset(bd,adr,m) (atomic_ops[bd]->OutB)(((atomic_ops[bd]->InB)((int) &((canregs_t *)Base[bd])->adr)) & ~m,(int) &((canregs_t *)Base[bd])->adr )
#	define CANtest(bd,adr,m) ((atomic_ops[bd]->InB)((int) &((canregs_t *)Base[bd])->adr  ) & m )
#endif

#if OLD_STYLE
#	define CANout(bd,adr,v) outb(v, (int) &((canregs_t *)Base[bd])->adr )
#	define CANin(bd,adr)    inb ((int) &((canregs_t *)Base[bd])->adr  )
#	define CANset(bd,adr,m) outb((inb((int) &((canregs_t *)Base[bd])->adr)) | m ,(int) &((canregs_t *)Base[bd])->adr )
#	define CANreset(bd,adr,m) outb((inb((int) &((canregs_t *)Base[bd])->adr)) & ~m,(int) &((canregs_t *)Base[bd])->adr )
#	define CANtest(bd,adr,m) (inb((int) &((canregs_t *)Base[bd])->adr  ) & m )
#endif
}/

/*======================================================================*/
file "82c200" /{

/* timing values */
uint8 CanTiming[9][2]={ {CAN_TIM0_10K,  CAN_TIM1_10K},
			{CAN_TIM0_20K,  CAN_TIM1_20K},
			{CAN_TIM0_40K,  CAN_TIM1_40K},
			{CAN_TIM0_50K,  CAN_TIM1_50K},
			{CAN_TIM0_125K, CAN_TIM1_125K},
			{CAN_TIM0_250K, CAN_TIM1_250K},
			{CAN_TIM0_500K, CAN_TIM1_500K},
			{CAN_TIM0_800K, CAN_TIM1_800K},
			{CAN_TIM0_1000K,CAN_TIM1_1000K}};

}/


/*======================================================================*/
header "82c200" /{
	
typedef unsigned char uint8;

extern uint8 CanTiming[9][2];

/*----------*/

typedef struct canregs {
	uint8    canctl;
	uint8    cancmd;
	uint8    canstat;
	uint8	 canirq;
	uint8	 canacc;
	uint8	 canmask;
	uint8	 cantim0;
	uint8	 cantim1;
	uint8	 canoutc;
	uint8	 cantest;
	uint8	 cantxdes1;
	uint8	 cantxdes2;
	uint8    cantxdata[8];
	uint8	 canrxdes1;
	uint8    canrxdes2;
	uint8    canrxdata[8];
	uint8    unused;
	uint8	 canclk; 	 
} canregs_t;

#define CAN_RANGE 0x20

/*--- Control Register ---------------------------------*/
 
#define CAN_TEST_MODE				(1<<7)
#define CAN_SPEED_MODE				(1<<6)
#define CAN_OVERRUN_INT_ENABLE			(1<<4)
#define CAN_ERROR_INT_ENABLE			(1<<3)
#define CAN_TRANSMIT_INT_ENABLE			(1<<2)
#define CAN_RECEIVE_INT_ENABLE			(1<<1)
#define CAN_RESET_REQUEST			(1<<0)

/*--- Command Register ------------------------------------*/
 
#define CAN_GOTO_SLEEP				(1<<4)
#define CAN_CLEAR_OVERRUN_STATUS		(1<<3)
#define CAN_RELEASE_RECEIVE_BUFFER		(1<<2)
#define CAN_ABORT_TRANSMISSION			(1<<1)
#define CAN_TRANSMISSION_REQUEST		(1<<0)

/*--- Status Register --------------------------------*/
 
#define CAN_BUS_STATUS 				(1<<7)
#define CAN_ERROR_STATUS			(1<<6)
#define CAN_TRANSMIT_STATUS			(1<<5)
#define CAN_RECEIVE_STATUS			(1<<4)
#define CAN_TRANSMISSION_COMPLETE_STATUS	(1<<3)
#define CAN_TRANSMIT_BUFFER_ACCESS		(1<<2)
#define CAN_DATA_OVERRUN			(1<<1)
#define CAN_RECEIVE_BUFFER_STATUS		(1<<0)

/*--- Status Register --------------------------------*/
 
#define CAN_BUS_STATUS_BIT 			(1<<7)
#define CAN_ERROR_STATUS_BIT			(1<<6)
#define CAN_TRANSMIT_STATUS_BIT			(1<<5)
#define CAN_RECEIVE_STATUS_BIT			(1<<4)
#define CAN_TRANSMISSION_COMPLETE_STATUS_BIT	(1<<3)
#define CAN_TRANSMIT_BUFFER_ACCESS_BIT		(1<<2)
#define CAN_DATA_OVERRUN_BIT			(1<<1)
#define CAN_RECEIVE_BUFFER_STATUS_BIT		(1<<0)

/*--- Interrupt Register -----------------------------------*/
 
#define CAN_WAKEUP_INT				(1<<4)
#define CAN_OVERRUN_INT				(1<<3)
#define CAN_ERROR_INT				(1<<2)
#define CAN_TRANSMIT_INT			(1<<1)
#define CAN_RECEIVE_INT 			(1<<0)

/*--- Output Control Register -----------------------------------------*/
/*
 *	7	6	5	4	3	2	1	0
 * 	OCTP1	OCTN1	OCPOL1	OCTP0	OCTN0	OCPOL0	OCMODE1	OCMODE0
 *	----------------------  ----------------------  ---------------
 *	    TX1 Output		    TX0 Output		  programmable
 *	  Driver Control	  Driver Control	  output functions
 *
 *	MODE
 *	OCMODE1	OCMODE0
 *	  0	  1	Normal Mode; TX0, TX1 bit sequenze TXData
 *	  1	  1	Normal Mode; TX0 bit sequenze, TX1 busclock TXCLK
 *	  0	  0	Biphase Mode
 *	  1	  0	Test Mode; TX0 bit sequenze, TX1 COMPOUT
 *
 *	In normal Mode Voltage Output Levels depend on 
 *	Driver Characteristic: OCTPx, OCTNx
 *	and programmed Output Polarity: OCPOLx
 *
 *	Driver Characteristic
 *	OCTPx	OCTNx
 *	  0	 0	always Floating Outputs,
 *	  0	 1	Pull Down
 *	  1	 0	Pull Up
 *	  1	 1	Push Pull
 */
 
/*--- Output control register --------------------------------*/

#define CAN_OCTP1			(1<<7)
#define CAN_OCTN1			(1<<6)
#define CAN_OCPOL1			(1<<5)
#define CAN_OCTP0			(1<<4)
#define CAN_OCTN0			(1<<3)
#define CAN_OCPOL0			(1<<2)
#define CAN_OCMODE1			(1<<1)
#define CAN_OCMODE0			(1<<0)

/*--- Remote Request ---------------------------------*/
/*    Notes:	RTR is Bit 5 in TXDES1.
 */
#define CAN_RTR					(1<<4)


/* the base address register array */
extern int Base[];
/*---------- Timing values */

/* the timings are valid for clock 8Mhz */
#define CAN_TIM0_10K		  49
#define CAN_TIM1_10K		0x1c
#define CAN_TIM0_20K		  24	
#define CAN_TIM1_20K		0x1c
#define CAN_TIM0_40K		0x89	/* Old Bit Timing Standard of port */
#define CAN_TIM1_40K		0xEB	/* Old Bit Timing Standard of port */
#define CAN_TIM0_50K		   9
#define CAN_TIM1_50K		0x1c
#define CAN_TIM0_125K		   3
#define CAN_TIM1_125K		0x1c
#define CAN_TIM0_250K		   1
#define CAN_TIM1_250K		0x1c
#define CAN_TIM0_500K		   0
#define CAN_TIM1_500K		0x1c
#define CAN_TIM0_800K		   0
#define CAN_TIM1_800K		0x16
#define CAN_TIM0_1000K		   0
#define CAN_TIM1_1000K		0x14


/*------------- CPC Specific registers -*/

extern volatile int CPC_Base;

#define cpc_contr  (CPC_Base)
#define cpc_status (CPC_Base+6)
#define cpc_type   (CPC_Base+4)


/*------------- PCCAN specific registers */

#define pccan_irqs   (pccan_Base | 0x6000 )
#define pccan_reset  (pccan_Base | 0x6001 )
#define pccan_isrc   (pccan_Base | 0x6002 )
#define PCCAN_CLOCK_INT  (1<<7)
#define PCCAN_CLOCK_BUS  0 
}/
/*======================================================================*/


file "82c200_irq" /{


static volatile int IRQbusy=0;
/*
 * The CPC interrupt wrapper
 */
void Can_CPC_Interrupt ( int irq, void *unused, struct pt_regs *ptregs ) {
   int minor = irq2minormap[irq];
   unsigned long flags;

   irq2minormap[0]=0;

#if CPCDEBUG
   printk("** CPC interrupt status=0x%x \n",readb(cpc_status));
#endif
   /*save_flags(flags);cli();restore_flags(flags);*/
   do {
       
       if( readb(cpc_status) & 0x80 ) irq2minormap[0] = 1;         /* this is an ugly trick I know... */
       if( readb(cpc_status) & 0x40 ) irq2minormap[0] = 0;         /* store the fixed channel in irq2minormap */
       Can_c200_Interrupt ( 0 , unused, ptregs );                  /* call the 82c200 interrupt */
       sti();
   } while ( readb(cpc_status) & 0xc0 );
}


/*
 *   the pccan interrupt wrapper
 */

void Can_pccan_Interrupt ( int irq, void *unused, struct pt_regs *ptregs ) {
   extern void Can_i527_Interrupt ( int irq, void *unused, struct pt_regs *ptregs );
   extern volatile int pccan_Base;
   int minor = irq2minormap[irq];
   int rirq=0,i;

   irq2minormap[0] = 0;
   rirq= ~ ( inb(pccan_isrc) & 0xf );
#if PCCAN_DEBUG
   printk("** PPCAN interrupt status=0x%x \n",rirq);
#endif
   do {
   
   for(i=0;i<=3;i++) {
     switch( rirq & (1<<i) ){
	case 1:
	   irq2minormap[0] = 2;
	   Can_i527_Interrupt ( 0 , unused, ptregs );
	   break;	
	case 2:
           irq2minormap[0] = 3;
	   Can_i527_Interrupt ( 0 , unused, ptregs );
	   break;
	case 4:
	   irq2minormap[0] = 0;
           Can_c200_Interrupt ( 0 , unused, ptregs );
           break;
	case 8:
	   irq2minormap[0] = 1;
           Can_c200_Interrupt ( 0 , unused, ptregs );
           break;
     }
   }

   rirq= ( ~ inb(pccan_isrc)) & 0xf ;

   } while (rirq);

#if 0
   do {
       if( rirq & 0x4 ) irq2minormap[0] = 0;           
       if( rirq & 0x8 ) irq2minormap[0] = 1;
       Can_c200_Interrupt ( 0 , unused, ptregs );
       if( rirq & 0x1 ) irq2minormap[0] = 2;
       if( rirq & 0x2 ) irq2minormap[0] = 3;
       Can_i527_Interrupt ( 0 , unused, ptregs );

       sti(); 
   } while ( (rirq= ~ (inb(pccan_isrc) & 0xf)) & 0xc );
#endif

}







/*
 * The plain 82c200 interrupt
 */

void Can_c200_Interrupt ( int irq, void *unused, struct pt_regs *ptregs ) {

extern volatile can_atomic_func_t *atomic_ops[];

   int minor = irq2minormap[irq];
   int i,j;
   unsigned long flags;
   int irqsrc,dummy;
   msg_fifo_t   *RxFifo = &Rx_Buf[minor]; 
   msg_fifo_t   *TxFifo = &Tx_Buf[minor];
   msg_filter_t *RxPass = &Rx_Filter[minor];
#if USE_FILTER
   unsigned int id;
#endif 
   canmsg_t msg;

   irqsrc = CANin(minor,canirq);

#ifdef IDEBUG   
   printk("CAN Interrupt %d (minor=%d irq=0x%x st=0x%x Rxhead=%d Rxtail=%d  Txhead=%d Txtail=%d )!\n",
           irq,minor,irqsrc,CANin(minor,canstat),RxFifo->head, RxFifo->tail,
           TxFifo->head, TxFifo->tail
           );
#endif
  
   /*========== receive interrupt */
   if( irqsrc & CAN_RECEIVE_INT ) {

        /*for(i=0;i<100 && (CANin(minor,canstat) & CAN_RECEIVE_STATUS );i++); */


        (RxFifo->data[RxFifo->head]).flags = (RxFifo->status & BUF_OVERRUN ? MSG_OVR : 0);

	dummy  = CANin(minor, canrxdes2 );
#if USE_FILTER
	id     = CANin(minor, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 ;

       	/* got RTR */
	if( dummy & CAN_RTR && RxPass->filter[id].rtr_response != NULL ) {
	        canmsg_t *RTRR = RxPass->filter[id].rtr_response;
#if IDEBUG
printk("got RTR %d!\n",RTRR->id);
#endif
		CANout( minor, cantxdes1 ,(uint8) (RTRR->id >> 3));
		CANout( minor, cantxdes2 ,(uint8) ((RTRR->id << 5) & 0xe0) | RTRR->length );
		for(i=0;i< RTRR->length && i<8 ; i++ )
		  CANout(minor, cantxdata[i], RTRR->data[i]); 
                CANout(minor, cancmd, CAN_TRANSMISSION_REQUEST );
	} else if( dummy & CAN_RTR ) {
                (RxFifo->data[RxFifo->head]).flags |= MSG_RTR;
        }

	(RxFifo->data[RxFifo->head]).id = id;
#else
        if( dummy & CAN_RTR ) 
              (RxFifo->data[RxFifo->head]).flags |= MSG_RTR;

        (RxFifo->data[RxFifo->head]).id = CANin(minor, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 ;
#endif

        dummy  &= 0x0f; /* strip length code */ 
        (RxFifo->data[RxFifo->head]).length = dummy;

        for( i=0;i<dummy && i<8 ;i++) {
           SLOW_DOWN_IO;SLOW_DOWN_IO;
	   (RxFifo->data[RxFifo->head]).data[i] = CANin(minor,canrxdata[i]);
#if IDEBUG
	printk("igot '%c' 0x%x irq=0x%x st=0x%x\n",CANin(minor,canrxdata[i]),(int) &((canregs_t *)Base[minor])->canrxdata[i], 
		CANin(minor,canirq),
		CANin(minor,canstat)  );
#endif
	}
	RxFifo->status = BUF_OK;
#if USE_FILTER
	if(RxPass->use && !RxPass->filter[id].enable){
#if FDEBUG
		printk("discarded %d \n",id);
#endif
		goto DiscardRX;
        }
	else
#endif
        RxFifo->head = ++(RxFifo->head) % MAX_BUFSIZE;

	if(RxFifo->head == RxFifo->tail) {
		printk("%d RX: FIFO overrun\n",minor);
		RxFifo->status = BUF_OVERRUN;
        } 
DiscardRX:
        /*---------- kick the select() call  -*/
        wake_up_interruptible(  &CanWait[minor] ); 

        CANout(minor, cancmd, CAN_RELEASE_RECEIVE_BUFFER );
        if(CANin(minor,canstat) & CAN_DATA_OVERRUN ) {
		 printk("Rx: Overrun Status \n");
		 CANout(minor, cancmd, CAN_CLEAR_OVERRUN_STATUS );
	}

goto IRQdone;
   }

   /*========== transmit interrupt */
   if( irqsrc & CAN_TRANSMIT_INT ) {
	uint8 tx2reg;
	if( TxFifo->head == TxFifo->tail  ) {
	   TxFifo->status = BUF_EMPTY;
           TxFifo->active = 0;
           goto Tx_done;
	}

        save_flags(flags);cli();

        CANout(minor, cantxdes1, (uint8)(( (TxFifo->data[TxFifo->tail]).id ) >> 3));
        tx2reg =  (( (TxFifo->data[TxFifo->tail]).id << 5 ) & 0xe0) 
                     | ((uint8) (TxFifo->data[TxFifo->tail]).length  ) ;
        if( (TxFifo->data[TxFifo->tail]).flags & MSG_RTR ) 
	    tx2reg |= CAN_RTR;
        CANout(minor, cantxdes2, tx2reg );

        for( i=0; i < (TxFifo->data[TxFifo->tail]).length && i<8 ; i++){ 

	CANout(minor, cantxdata[i], (TxFifo->data[TxFifo->tail]).data[i]);
#if 0
printk("isend '%c' 0x%x \n ", CANin(minor, cantxdata[i]),&((canregs_t *)Base[minor])->cantxdata[i] );
#endif 
         SLOW_DOWN_IO;
	}

        CANout(minor, cancmd, CAN_TRANSMISSION_REQUEST );
        for( i=0;i<1000 && ! (CANin(minor,canstat) & CAN_TRANSMISSION_COMPLETE_STATUS ) ;i++);
#if 0
printk("%d: CANstat=0x%x j=%x cycle=%d \n",minor,CANin(minor,canstat),j,i);
#endif
	TxFifo->tail = ++(TxFifo->tail) % MAX_BUFSIZE;

	restore_flags(flags);

	goto IRQdone;
   }
Tx_done:
   /*========== error status */
   if( irqsrc & CAN_ERROR_INT ) {
	printk("%d: Tx err!\n",minor);
        TxErr[minor]++;
	
   }
   if( irqsrc & CAN_OVERRUN_INT ) {
	printk("%d: overrun!\n",minor);
        Overrun[minor]++;
        CANout(minor, cancmd, CAN_CLEAR_OVERRUN_STATUS );
   }

IRQdone:
}


}/


/*======================================================================*/

file "82c200funcs" /{


extern int Can_c200_ChipReset (int);
extern int Can_c200_SetTiming (int,int);
extern int Can_c200_StartChip(int);
extern int Can_c200_StopChip(int);
extern int Can_c200_SetMask(int,int);
extern int Can_c200_SetOMode(int,int);
extern int Can_c200_SendMessage(int, canmsg_t *, int );
extern int Can_c200_GetMessage(int , canmsg_t *);
extern void Can_c200_Interrupt( int irq, void *unused, struct pt_regs *ptregs );
extern int Can_c200_VendorInit (int);

can_atomic_func_t p82c200_ops = {
   Can_c200_ChipReset,
   Can_c200_SetTiming,
   Can_c200_StartChip,
   Can_c200_StopChip,
   Can_c200_SetMask,
   Can_c200_SetOMode,
   Can_c200_SendMessage,
   Can_c200_GetMessage,
   Can_c200_Interrupt,
   Can_c200_VendorInit,
#if 0
   PORT_In,
   PORT_Out
#endif
};

extern int Can_CPC_VendorInit (int);
extern void Can_CPC_Interrupt( int irq, void *unused, struct pt_regs *ptregs );

can_atomic_func_t emscpc_ops = {
   Can_c200_ChipReset,
   Can_c200_SetTiming,
   Can_c200_StartChip,
   Can_c200_StopChip,
   Can_c200_SetMask,
   Can_c200_SetOMode,
   Can_c200_SendMessage,
   Can_c200_GetMessage,
   Can_CPC_Interrupt,
   Can_CPC_VendorInit,
#if 0
   MEM_In,
   MEM_Out
#endif
};

extern int Can_pccan_VendorInit (int);
extern void Can_pccan_Interrupt( int irq, void *unused, struct pt_regs *ptregs );

can_atomic_func_t pccan_ops = {
   Can_c200_ChipReset,
   Can_c200_SetTiming,
   Can_c200_StartChip,
   Can_c200_StopChip,
   Can_c200_SetMask,
   Can_c200_SetOMode,
   Can_c200_SendMessage,
   Can_c200_GetMessage,
   Can_pccan_Interrupt,
   Can_pccan_VendorInit,
#if 0
   PORT_In,
   PORT_Out
#endif
};



int          CAN_Open = 0;


/* global variables for CPC */

volatile int CPC_Base = 0;

/* global variables for pccan */

volatile int pccan_Base = 0;


}/

/@ performs vendor specific initialization for the plain 82c200 @/

func c200_VendorInit (int minor) /{

      /*extern void Can_Interrupt( int, void *, struct pt_regs *);*/

      if( check_region(Base[minor],CAN_RANGE ) ){
		     MOD_DEC_USE_COUNT;
		     DBGout();
		     return -EBUSY;
      }
       else
          request_region(Base[minor],CAN_RANGE );

      if( Outc[minor] == 0 ) {
	Outc[minor] = CAN_OUTC_VAL;
      }
      if( VendOpt[minor] == 's' ) {    /* stpz board */
        printk("Resetting STZP PC I-03\n");
	(io_funcs[minor].OutB)(0xff, Base[minor] + 0x100 ); /* perform HW reset 1. contr*/
	(io_funcs[minor].OutB)(0xff, Base[minor] + 0x300 ); /* perform HW reset 2. contr*/
      }
      if( IRQ[minor] > 0 ) {
	       if( Can_RequestIrq( minor, IRQ[minor] , 
                         CAN_Interrupt(minor) ) ){
			     printk("Can[%d]: Can't request IRQ %d \n",minor, IRQ[minor] );
			     DBGout(); return -EBUSY;
	       }
      }
      DBGout(); return 1;
 }/


/@ performs vendor specific initialization for the kvaser ab pccan adapter @/

func pccan_VendorInit (int minor) /{
      extern int CAN_Open;
      extern volatile int pccan_Base;
      int istr_val=0;
      int i;

      if(!CAN_Open) {
        pccan_Base = Base[0] & 0xfff;
      
        Base[0] = pccan_Base | 0x4000;      /* can0/1 are the phillips chips */
        Base[1] = Base[0] | 0x5000;
        Base[2] = Base[0] | 0x2000;         /* intel 1 */
        Base[3] = Base[0] | 0x3000;

#if 0
        if( check_region( pccan_Base ,CAN_RANGE ) ){
                     printk("Can[%d]: Problem getting region \n", minor );
		     MOD_DEC_USE_COUNT;
		     DBGout();
		     return -EBUSY;
        }
        else
        request_region( pccan_Base ,CAN_RANGE );
#endif

        for(i=0;i<4;i++) {
           if( Outc[i] == 0 ) {
	             Outc[i] = PCCAN_OUTC_VAL;
           }
           IRQ[i] = IRQ[0];
           Chipset[i] = Chipset[0];
           IOModel[i] = 'p';
        }

        /* chip hardware reset via Xlinx */

        outb(0x0, pccan_reset);
        for(i=0;i<1000;i++) SLOW_DOWN_IO;
	outb(0x1, pccan_reset);

        if( IRQ[minor] > 0 ) {
	       if( Can_RequestIrq( minor, IRQ[minor] , 
                         CAN_Interrupt(minor) ) ){
			     printk("Can[%d]: Can't request IRQ %d \n",minor, IRQ[minor] );
			     DBGout(); return -EBUSY;
	       }
        }

        /* setting interrupt, in this version all int sources are routed to one routine */

        if( IRQ[0] == 2 ) istr_val = 0x5555; 
        if( IRQ[0] == 3 ) istr_val = 0xaaaa; 
        if( IRQ[0] == 5 ) istr_val = 0xffff; 
        outb(istr_val , pccan_irqs);

        /* select clock frequency PCCAN_CLOCK_INT=internal PCCAN_CLOCK_BUS=bus clock */

        outb( pccan_isrc, PCCAN_CLOCK_INT );


      }
    CAN_Open++;
    DBGout(); return 1;
}/




/@ performs vendor specific initialization for the EMS CPC card @/
func CPC_VendorInit (int minor) /{
 int i,found=0;
 extern int CAN_Open;
 extern volatile int CPC_Base;

 /* check for card */ 

 if( minor > 1 ){
      printk(" ** CPC is only supported on port 0/1 ! \n");
      DBGout(); 
      return -EINVAL;
 }
 if(! CAN_Open ) {
    for(i=0xc0000;i<0xe0000;i+=0x20){
	if( readb(i)   == 0x55 &&
	    readb(i+1) == 0xaa &&
	    readb(i+2) == 0x01 &&
            readb(i+3) == 0xcb ) {
  		found = 1;
		break;
        }
   }
   if (found) {
                if( IRQ[0] == 0 || IRQ[1] == 0 ) { 
		   IRQ[0]    = 5;
		   IRQ[1]    = 5;
                }

                if( Outc[minor] == 0 ) {
		   Outc[0] = Outc[1] = CPC_OUTC_VAL;
                }
                CPC_Base  = i;
		Base[0]   = i+0x40;
		Base[1]   = i+0x60;
		Chipset[0] = CHIPSET_emscpc;
		Chipset[1] = CHIPSET_emscpc;
                Set_IO_Model(0,'m');
                Set_IO_Model(1,'m');
		
		printk("  ** EMS-CPC card type %d found at 0x%x \n",readb(CPC_Base+4),i);
	        if( Can_RequestIrq( minor, IRQ[minor] , 
                         CAN_Interrupt(minor) ) ){
			     printk("Can[%d]: Can't request IRQ %d \n",minor, IRQ[minor] );
			     DBGout(); return -EBUSY;
	        }
                writeb(0, cpc_contr );  /* perform hardware reset */
                /*printk("  ** CPC status=0x%x\n",readb(cpc_status) );*/
   } else {
     DBGout();
     return -EINVAL;
   }
 }
 CAN_Open++;
 DBGout(); 
 return 1;

}/


/@
  Resetting 82C200 Controller
  board is the board id
@/
func c200_ChipReset (int board) /{

	uint8 status;
	int i;

	printk("Resetting Board %d on 0x%x outc=0x%x \n", board, Base[board],Outc[board]);
	DBGprint(DBG_DATA,("cmdr=0x%x\n", &((canregs_t *)Base[board])->cancmd ));

	CANset(board, canctl, CAN_RESET_REQUEST );
	for(i=0;i<100;i++) SLOW_DOWN_IO;

	status = CANin(board, canstat);
	DBGprint(DBG_DATA,("status=0x%x ctl=0x%x",status,
		CANin(board, canctl) ));
	if( ! (CANin(board, canctl) & CAN_RESET_REQUEST ) ){
		printk("ERROR: no board present!");
                MOD_DEC_USE_COUNT;
		DBGout();return -1;
	}
	
        /* Board specific output control */

        CANout(board, canoutc, Outc[board]);

	Can_c200_SetTiming(board, Baud[board]    );
	Can_c200_SetMask  (board, AccMask[board] );

	DBGout();
	return status;
}/


/@
  Set timing values
@/
func c200_SetTiming (int board,int baud) /{
int i; int custom=0;
    DBGprint(DBG_DATA,("baud[%d]=%d",board,baud));
    switch(baud)
    {
	case   10: i = 0; break;
	case   20: i = 1; break;
	case   40: i = 2; break;
	case   50: i = 3; break;
	case  125: i = 4; break;
	case  250: i = 5; break;
	case  500: i = 6; break;
	case  800: i = 7; break;
	case 1000: i = 8; break;
	default  : 
		custom=1;
    }
    CANout(board,canclk,0);
    if( custom ) {
       CANout(board,cantim0, (uint8) (baud >> 8) & 0xff);
       CANout(board,cantim1, (uint8) (baud & 0xff ));
       printk(" custom bit timing BT0=0x%x BT1=0x%x ",CANin( board,cantim0),CANin( board,cantim1));
    } else {
       CANout(board,cantim0, (uint8) CanTiming[i][0]);
       CANout(board,cantim1, (uint8) CanTiming[i][1]);
    }
    DBGprint(DBG_DATA,("tim0=0x%x tim1=0x%x",CANin( board,cantim0),CANin( board,cantim1)) );

}/
/*======================================================================*/
/@
 Start Controller
@/
func c200_StartChip (int board) /{
	int i;
	RxErr[board] = TxErr[board] = 0L;
/*
	CANout( board,cancmd, (CAN_RELEASE_RECEIVE_BUFFER 
                              | CAN_CLEAR_OVERRUN_STATUS) ); 
*/
	CANreset( board, canctl, CAN_RESET_REQUEST );
	for(i=0;i<100;i++) SLOW_DOWN_IO;
	CANin(board,canirq);

CANset( board, canctl, CAN_RECEIVE_INT_ENABLE );
	DBGprint(DBG_DATA,("start ctl=0x%x",CANin( board,canctl)) );
}/

/@ 
 Stop Controller 
 @/
func c200_StopChip (int board) /{
	CANset(board, canctl, CAN_RESET_REQUEST );
}/

/@ 
 Set ACCMASK and ACCR
 @/
func c200_SetMask (int board, int arg) /{
	DBGprint(DBG_DATA,("acc=0x%x mask=0x%x", 
			  (arg & 0xffff) >> 8,(arg & 0xff)  ));
	CANout(board, canacc, (arg & 0xffff) >> 8 );	
	CANout(board, canmask, (arg & 0xff)  );
}/

/@ 
 Set OUTC register
 @/
func c200_SetOMode (int board, int arg) /{
	DBGprint(DBG_DATA,("outc=0x%x",arg));
	CANout(board,canoutc, arg );
}/

/@

 Send CAN Message to Tx Buffer with polling

 @/
func c200_SendMessage(int board, canmsg_t *tx, int wait) /{
  int i=0,cycle=0,k=0;
  uint8 tx2reg,stat;


  if(wait) LDDK_START_TIMER(Timeout[board]);
  while ( ! (stat=CANin(board,canstat)) & CAN_TRANSMIT_BUFFER_ACCESS && ! LDDK_TIMEDOUT ) {
	if( need_resched ) schedule();
  }

  if( ! LDDK_TIMEDOUT  ) {
    DBGprint(DBG_DATA,("0x%x: tx.flags=%d tx.id=0x%x tx.length=%d stat=0x%x",
                   Base[board],tx->flags,tx->id,tx->length,stat));

    CANout(board, cantxdes1, (uint8)((tx->id) >> 3) );
    tx2reg =  ((tx->id << 5 ) & 0xe0) | ((uint8)(tx->length)) ;
    if( tx->flags & MSG_RTR ) 
	    tx2reg |= CAN_RTR;
    CANout(board, cantxdes2, tx2reg );

    for( i=0; i < tx->length ; i++) {
	CANout( board, cantxdata[i], tx->data[i]); SLOW_DOWN_IO; SLOW_DOWN_IO;
#if 0
printk("send '%c' 0x%x\n",CANin(board, cantxdata[i]), (int) &((canregs_t *)Base[board])->cantxdata[i] );
	DBGprint(DBG_DATA,("tx[%d]: 0x%x ('%c')",i,tx->data[i],tx->data[i]));
#endif
    }
    CANout(board, cancmd, CAN_TRANSMISSION_REQUEST );
    if( wait ) {
       while( ! (CANin(board, canstat) & CAN_TRANSMISSION_COMPLETE_STATUS) 
           && ! LDDK_TIMEDOUT ) cycle++;
       if( LDDK_TIMEDOUT ) {
       DBGprint(DBG_DATA,("Timeout! stat=0x%x",CANin(board, canstat  )));	
	/*CANout(board, cancmd, CAN_ABORT_TRANSMISSION );*/
    

       /*CANout( board,cancmd, (CAN_RELEASE_RECEIVE_BUFFER 
	                              | CAN_CLEAR_OVERRUN_STATUS) );*/
       }
    } else if( TxSpeed[board] == 'f' ) {
      unsigned long flags;
      save_flags(flags); cli();
      for(i=0;i<1000 && !( CANin(board, canstat) & CAN_TRANSMISSION_COMPLETE_STATUS ) ;i++);
      restore_flags(flags);
    }
  }  else {
    printk("Tx: Timeout! stat=0x%x\n",stat);
    i= -1;
  }

  if(wait) LDDK_STOP_TIMER();
  DBGout();return i;
}/


/@
 Get Message from Receive Buffer with polling
@/
func c200_GetMessage(int board, canmsg_t *rx ) /{
  uint8 dummy,stat;
  int i=0;
  
  LDDK_START_TIMER(Timeout[board]);
  stat = CANin(board, canstat  );
  DBGprint(DBG_DATA,("0x%x:stat=0x%x ctl=0x%x",Base[board],stat,CANin(board,canctl)));
  rx->flags=0;
  rx->length=0;

  if( stat & CAN_DATA_OVERRUN ) {
	DBGprint(DBG_DATA,("Rx: overrun!\n"));
	Overrun[board]++;
  }

  if( stat & CAN_RECEIVE_BUFFER_STATUS ) {
    dummy  = CANin(board, canrxdes2 );
    rx->id = CANin(board, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 ;
    dummy  &= 0x0f; /* strip length code */ 
    rx->length = dummy;
    DBGprint(DBG_DATA,("rx.id=0x%x rx.length=0x%x",rx->id,dummy));

    for( i=0;i<dummy && i<8 ;i++) {
	rx->data[i] = CANin(board,canrxdata[i]);
	DBGprint(DBG_DATA,("rx[%d]: 0x%x ('%c')",i,rx->data[i],rx->data[i] ));
    }
    CANout(board, cancmd, CAN_RELEASE_RECEIVE_BUFFER | CAN_CLEAR_OVERRUN_STATUS );
  } else {
    i=0;
  }
  LDDK_STOP_TIMER();
  DBGout();
  return i;
}/


