library {

header "lib_config" /{
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

typedef struct {
    char base[40];
    char baud[40];
    char irq[40];
    char accmask[40];
    char timeout[40];
} can_conf_t;

#define ERR -1
#define CONFIG_FILE "/etc/can.conf"

}/

file "lib_config" /{
 
can_conf_t LibConfig;

}/

/@
   Initializes the Can Subsystem, port is the can port number (0-3)
@/

func CanInit ( int port ) /{
	int handle;
	char inode[50];
	
	sprintf(inode,"/dev/port%d",port);	
        if(( handle = open( inode, O_RDWR )) < 0 ) {
	  fprintf(stderr,"CANLIB:Can't open '%s'\n",inode );
	  return ERR;
        }

	return handle;
}/

/@
   Deinitializes Can Subsystem 
@/

func CanRelease ( int handle ) /{
	if(handle)
		close(handle);
}/


func CanLoadConfig() /{
  FILE *fp;

  if(( fp = fopen(CONFIG_FILE,"rb")) == NULL) {
	fprintf(stderr,"CANLIB: Can't open Config File\n");
	return ERR;
  }
  if( fread( &LibConfig, sizeof(can_config_t),1,fp ) < 1 )
	fprintf(stderr,"CANLIB: Can't read Config File\n");
	return ERR;
  }
  fclose(fp);
  return 1;	


}/

func CanWriteConfig() /{
  FILE *fp;

  if(( fp = fopen(CONFIG_FILE,"wb")) == NULL ) {
    Error("CANLIB: Can't open Configuration File for reading\n");
    return (0);
  }
  if( fwrite( &LibConfig,sizeof(can_config_t),1,fp) < 1 ) {
    Error("CANLIB:Error Writing Config\n");
    fclose(fp);
    return(0);
  }
  fclose(fp);  
  return(1);    
}/


}