/*
 *  Support file for the Intel i82527 CAN controller
 *
 *  - currently tested for EMS CPC-XT
 *  - support for STZP PC-I 03
 *
 */

header "i527_conf" /{
  /*
   * default configuration
   *
   */
#define IODEBUG 0
}/

header "i527_regs" /{
#include <linux/delay.h>

/* Register offset mappings in the register map */

#	define icanctrl 	0x0
#	define istatus  	0x1
#	define icpuif   	0x2
#	define ires0    	0x3
#	define ihsr0    	0x4
#	define ihsr1    	0x5
#	define iglm_s(i)  	(0x6+i)
#	define iglm_e(i)  	(0x8+i)
#	define im15mask(i) 	(0xC+i)
#	define iclkout		0x1f
#	define ibuscfg		0x2f
#	define icantim0        	0x3f
#	define icantim1        	0x4f
#	define iintreg		0x5f
#	define ip1conf		0x9f
#	define ip2conf		0xaf
#	define ip1in		0xbf
#	define ip2in		0xcf
#	define ip1out		0xdf
#	define ip2out		0xef
#	define iserres         	0xff

#	define iMsgReg(a,b)    ((a<<4)+b)

/* register values  */

/* control register */
#	define iCAN_CCE   	(1<<6)
#	define iCAN_EIE        	(1<<3)
#	define iCAN_SIE		(1<<2)
#	define iCAN_IE		(1<<1)
#	define iCAN_INIT	(1<<0)
/* status register */
#	define iCAN_BOFF	(1<<7)
#	define iCAN_WARN	(1<<6)
#	define iCAN_WAKE	(1<<5)
#	define iCAN_RXOK	(1<<4)
#	define iCAN_TXOK	(1<<3)
#	define iCAN_LEC2	(1<<2)
#	define iCAN_LEC1	(1<<1)
#	define iCAN_LEC0	(1<<0)
/* cpu interface register */
#	define iCAN_RSTST	(1<<7)
#	define iCAN_DSC		(1<<6)
#	define iCAN_DMC		(1<<5)
#	define iCAN_PWD		(1<<4)
#	define iCAN_SLEEP	(1<<3)
#	define iCAN_MUX		(1<<2)
#	define iCAN_CEN		(1<<0)

/* message control */


/* message control register */

#define MSGVAL_RST 0x40
#define MSGVAL_SET 0x80
#define MSGVAL_NOP 0xC0

#define TXIE_RST   0x10
#define TXIE_SET   0x20
#define TXIE_NOP   0x30

#define RXIE_RST   0x04
#define RXIE_SET   0x08
#define RXIE_NOP   0x0C

#define INTPND_RST 0x01
#define INTPND_SET 0x02
#define INTPND_NOP 0x03

#define RMTPND_RST 0x40
#define RMTPMD_SET 0x80
#define RMTPND_NOP 0xC0

#define TXRQST_RST 0x10
#define TXRQST_SET 0x20
#define TXRQST_NOP 0x30

#define CPUUPD_RST 0x04
#define CPUUPD_SET 0x08
#define CPUUPD_NOP 0x0C

#define MSGLST_RST 0x04
#define MSGLST_SET 0x08
#define MSGLST_NOP 0x0C

#define NEWDAT_RST 0x01
#define NEWDAT_SET 0x02
#define NEWDAT_NOP 0x03

}/

file "i527io" /{

unsigned long i527_AddrMap[4][257];

}/ 

header "i527io" /{

extern unsigned long i527_AddrMap[4][257];
    
#	define iCANout(bd,adr,v)  (io_funcs[bd].OutB)(v,(i527_AddrMap[bd][adr]))
#	define iCANin(bd,adr)     (io_funcs[bd].InB)(i527_AddrMap[bd][adr])

}/


file "i527_irq" /{

void Can_i527_Interrupt ( int irq, void *unused, struct pt_regs *ptregs ) {

   extern volatile can_atomic_func_t *atomic_ops[];

   int minor = irq2minormap[irq];
   int i,j;
   unsigned long flags;
   int irqsrc,dummy,istat;
   msg_fifo_t   *RxFifo = &Rx_Buf[minor]; 
   msg_fifo_t   *TxFifo = &Tx_Buf[minor];
   msg_filter_t *RxPass = &Rx_Filter[minor];

   unsigned int cobid,format;

   canmsg_t msg;
   
   cobid = iCANin(minor,iintreg); 
   istat = iCANin(minor,istatus);
   printk("iCAN Interrupt %d (minor=%d cobid=%d irq=0x%x st=0x%x Rxhead=%d Rxtail=%d  Txhead=%d Txtail=%d )!\n",
           irq,minor,cobid,irqsrc,istat,RxFifo->head, RxFifo->tail,
           TxFifo->head, TxFifo->tail
           );

   if( cobid == 2 ) { /* message through COB 15 */

        (RxFifo->data[RxFifo->head]).flags = (RxFifo->status & BUF_OVERRUN ? MSG_OVR : 0);

        dummy = iCANin(minor,iMsgReg(15,6)); /* get cfg */
        if( dummy & 0x4 ) { /* xtd bit set ? */
             (RxFifo->data[RxFifo->head]).id = 
					      (iCANin(minor,iMsgReg(15,2)) & 0xff) << 21	
					    | (iCANin(minor,iMsgReg(15,3)) & 0xff) << 13	
                                            | (iCANin(minor,iMsgReg(15,4)) & 0xff) << 5 
                                            | (iCANin(minor,iMsgReg(15,5)) & 0xf8) >> 3 ;       
        } else { 
          (RxFifo->data[RxFifo->head]).id = iCANin(minor,iMsgReg(15,2)) << 3 
                                            | (iCANin(minor,iMsgReg(15,3)) & 0xe0) >> 5 ;
         
        }
        (RxFifo->data[RxFifo->head]).length = (dummy & 0xf0) >> 4;
#if IDEBUG
         printk("id=0x%x length=0x%d \n ",(RxFifo->data[RxFifo->head]).id,(RxFifo->data[RxFifo->head]).length);
#endif
        for( i=0;i<(RxFifo->data[RxFifo->head]).length && i<8 ;i++) {
	   (RxFifo->data[RxFifo->head]).data[i] = iCANin(minor,iMsgReg(15, 7+i ));
#if IDEBUG
	   printk("igot[%d] '%c' 0x%x \n",i,(RxFifo->data[RxFifo->head]).data[i],(RxFifo->data[RxFifo->head]).data[i] );
#endif
	}

	RxFifo->status = BUF_OK;
        RxFifo->head = ++(RxFifo->head) % MAX_BUFSIZE;

	if(RxFifo->head == RxFifo->tail) {
		printk("%d RX: FIFO overrun\n",minor);
		RxFifo->status = BUF_OVERRUN;
        } 


	iCANout(minor,iMsgReg(15,0), MSGVAL_NOP | TXIE_NOP   | RXIE_NOP   | INTPND_RST );
	iCANout(minor,iMsgReg(15,1), RMTPND_RST | TXRQST_NOP | CPUUPD_NOP | NEWDAT_RST );

        /*---------- kick the select() call  -*/
        wake_up_interruptible(  &CanWait[minor] ); 
   } 
   else
   if( cobid > 2 ) {
	/* clear Irq-pending bin in message object */
	iCANout(minor,iMsgReg(cobid-2,0), MSGVAL_NOP | TXIE_NOP   | RXIE_NOP   | INTPND_RST );
   } else {
        printk("Can[%d]: Status change IRQ sr=0x%x \n",minor,istat );
   }

}

}/



file "i527funcs" /{

extern int Can_i527_ChipReset (int);
extern int Can_i527_SetTiming (int,int);
extern int Can_i527_StartChip(int);
extern int Can_i527_StopChip(int);
extern int Can_i527_SetMask(int,int);
extern int Can_i527_SetOMode(int,int);
extern int Can_i527_SendMessage(int, canmsg_t *, int );
extern int Can_i527_GetMessage(int , canmsg_t *);
extern void Can_i527_Interrupt( int irq, void *unused, struct pt_regs *ptregs );
extern int Can_i527_VendorInit (int);

can_atomic_func_t i527_ops = {
   Can_i527_ChipReset,
   Can_i527_SetTiming,
   Can_i527_StartChip,
   Can_i527_StopChip,
   Can_i527_SetMask,
   Can_i527_SetOMode,
   Can_i527_SendMessage,
   Can_i527_GetMessage,
   Can_i527_Interrupt,
   Can_i527_VendorInit,
#if 0
   PORT_In,
   PORT_Out
#endif
};

}/


func i527_VendorInit (int minor) /{

    int i;
    if( VendOpt[minor] == 's' ) {    /* stzp board */
        printk("Can[%d]: Resetting STZP PC I-03 at 0x%x \n",minor,Base[minor]);
	(io_funcs[minor].OutB)(0xff, Base[minor] + 0x100 ); /* perform HW reset 1. contr*/
	(io_funcs[minor].OutB)(0xff, Base[minor] + 0x300 ); /* perform HW reset 2. contr*/

          for(i=0;i<0xff;i++) {
	     i527_AddrMap[minor][i] = i+Base[minor];
         }

    } else 
    if( VendOpt[minor] == 'e' ){
      /* try to detect CPC-XT board */
      if(   (io_funcs[minor].InB)(Base[minor]  ) != 0x55 
         && (io_funcs[minor].InB)(Base[minor]+1) != 0xaa
         && (io_funcs[minor].InB)(Base[minor]+3) != 0xcb ) {
	   printk("Can[%d]: not a EMS board at 0x%x \n",minor,Base[minor]);
	   DBGout(); return -EINVAL;
         }	
         printk("Can[%d]: EMS CPC-XT i82527 found at 0x%x \n",minor,Base[minor]);
         (io_funcs[minor].OutB)( 0x0, Base[minor] + 0 );

         for(i=0;i<0xff;i++) {
	     i527_AddrMap[minor][i] = i+Base[minor]+0x100;
         }
         (io_funcs[minor].OutB)( 0x3, Base[minor] + 0 );
    } else {
          printk("Can[%d]: Generic i82527 at 0x%x \n",minor,Base[minor]);
          for(i=0;i<0xff;i++) {
	     i527_AddrMap[minor][i] = i+Base[minor];
         }
    }

    if( IRQ[minor] > 0 ) {
	       if( Can_RequestIrq( minor, IRQ[minor] , 
                         CAN_Interrupt(minor) ) ){
			     printk("Can[%d]: Can't request IRQ %d \n",minor, IRQ[minor] );
			     DBGout(); return -EBUSY;
	       }
    }

    DBGout(); return 1;
}/

func i527_ChipReset (int minor) /{
	int i;
	(io_funcs[minor].OutB)( 0x0, Base[minor] + 0 );
	udelay( 100000 );	
	(io_funcs[minor].OutB)( 0x3, Base[minor] + 0 );

	for(i=0;i<=100 && (iCANin(minor,icpuif) & iCAN_RSTST  );i++);
        if( i>= 100) {
		printk("Can[%d]: i82527 did not reset! \n",minor);
		DBGout(); return -EINVAL;
	}

        iCANout(minor,icanctrl, 0x41 ); /* enable register access */

	iCANout(minor,ip2conf,  0x00 ); 	
	iCANout(minor,ibuscfg, Outc[minor] );     /* set bus configuration */
        DBGprint(DBG_DATA,("busconf=0x%x ",Outc[minor] ));

	Can_i527_SetTiming(minor, Baud[minor] );  /* set bus timing        */      
	Can_i527_SetMask  (minor, AccMask[minor] );

	iCANout(minor,iglm_s(0), 0x00 ); /* clear standard mask */
	iCANout(minor,iglm_s(1), 0x00 ); 

	iCANout(minor,iglm_e(0), 0x00 ); /* clear extended mask */
	iCANout(minor,iglm_e(1), 0x00 ); 
	iCANout(minor,iglm_e(2), 0x00 ); 
	iCANout(minor,iglm_e(3), 0x00 ); 

	for(i=1;i<=14;i++) {               /* reset all message control registers */
	  iCANout(minor,iMsgReg(i, 0 ),  MSGVAL_RST | TXIE_RST   | RXIE_RST   | INTPND_RST );
          iCANout(minor,iMsgReg(i, 1 ),  RMTPND_RST | TXRQST_RST | CPUUPD_RST | NEWDAT_RST );
	}


        /* configure message 15 for receive standard messages */
        iCANout(minor,iMsgReg(15, 0 ),  MSGVAL_SET | TXIE_NOP   | RXIE_SET   | INTPND_RST );
        iCANout(minor,iMsgReg(15, 1 ),  RMTPND_RST | TXRQST_NOP | MSGLST_RST | NEWDAT_RST );
        iCANout(minor,iMsgReg(15, 6 ),  0x0 ); /* direction = receive,XTD=0 */
	iCANout(minor,im15mask(0), 0x00 ); /* set mask 15 all bits valid */
	iCANout(minor,im15mask(1), 0x00 );  
	iCANout(minor,im15mask(2), 0x00 );  
	iCANout(minor,im15mask(3), 0x00 );  


        /* start_chip enables interrupts */

	DBGout();
	return 1;     
}/

func i527_SetTiming (int minor,int baud) /{
int i; int custom=0;
    DBGprint(DBG_DATA,("baud[%d]=%d",minor,baud));
    switch(baud)
    {
	case   10: i = 0; break;
	case   20: i = 1; break;
	case   40: i = 2; break;
	case   50: i = 3; break;
	case  125: i = 4; break;
	case  250: i = 5; break;
	case  500: i = 6; break;
	case  800: i = 7; break;
	case 1000: i = 8; break;
	default  : 
		custom=1;
    }
    CANout(minor,canclk,0);
    if( custom ) {
       iCANout(minor,icantim0, (uint8) (baud >> 8) & 0xff);
       iCANout(minor,icantim1, (uint8) (baud & 0xff ));
       printk(" custom bit timing BT0=0x%x BT1=0x%x ",CANin( minor,cantim0),CANin( minor,cantim1));
    } else {
       iCANout(minor,icantim0, (uint8) CanTiming[i][0]);
       iCANout(minor,icantim1, (uint8) CanTiming[i][1]);
    }
    DBGprint(DBG_DATA,("tim0=0x%x tim1=0x%x",CANin( minor,cantim0),CANin( minor,cantim1)) );
}/

func i527_StartChip (int minor) /{
   RxErr[minor] = TxErr[minor] = 0L;
   iCANout(minor,istatus , 0x00 );
   iCANout(minor,icanctrl, iCAN_EIE | iCAN_IE );
}/

func i527_StopChip (int minor) /{
   /*iCANout(minor,istatus , 0x00 )*/
   iCANout(minor,icanctrl, iCAN_SIE | iCAN_INIT );
}/


func i527_SetMask (int minor, int arg) /{
DBGprint(DBG_DATA,("mask=0x%x", arg ));
   iCANout(minor,iglm_s(0), (uint8) (arg >> 8) & 0xff);
   iCANout(minor,iglm_s(1), (uint8) (arg) & 0xff);
}/

func i527_SetXTDMask (int minor, unsigned long arg) /{
DBGprint(DBG_DATA,("mask=0x%lx",arg ));
   iCANout(minor,iglm_e(0), (uint8) (arg >> 24) & 0xff);
   iCANout(minor,iglm_e(1), (uint8) (arg >> 16) & 0xff);
   iCANout(minor,iglm_e(2), (uint8) (arg >> 8) & 0xff);
   iCANout(minor,iglm_e(3), (uint8) (arg) & 0xff);
}/


func i527_SetOMode (int minor, int arg) /{
	DBGprint(DBG_DATA,("outc=0x%x",arg));
	iCANout(minor,ibuscfg, arg );   
}/

func i527_SendMessage(int minor, canmsg_t *tx, int wait) /{
  int i,xtd;

  xtd = ( tx->flags & MSG_EXT ? 1 : 0 ); 

  if( tx->cob == 0) tx->cob=1;  /* fallback to cob=1 in apps that do not set cob */

  DBGprint(DBG_DATA,("Can[%d]: tx.flags=%d tx.cob=%d tx.id=0x%x tx.length=%d ",
                          minor,tx->flags,tx->cob,tx->id,tx->length));


  if(wait) LDDK_START_TIMER(Timeout[minor]);
  
  iCANout(minor,iMsgReg(tx->cob, 0 ),  MSGVAL_NOP | TXIE_NOP   | RXIE_NOP   | INTPND_NOP );
  iCANout(minor,iMsgReg(tx->cob, 1 ),  RMTPND_NOP | TXRQST_NOP | CPUUPD_SET | NEWDAT_SET );

  if( xtd ) { /* long ID format */ 
  	DBGprint(DBG_DATA,("Can[%d]: XTD",minor));
	iCANout(minor,iMsgReg(tx->cob, 2 ) , (tx->id >> 21)  );
	iCANout(minor,iMsgReg(tx->cob, 3 ) , (tx->id >> 13) & 0xff );
	iCANout(minor,iMsgReg(tx->cob, 4 ) , (tx->id >> 5 ) & 0xff );
	iCANout(minor,iMsgReg(tx->cob, 5 ) , (tx->id << 3 ) & 0xff );
  } else {
	iCANout(minor,iMsgReg(tx->cob, 2 ) , (tx->id >> 3 )  );
	iCANout(minor,iMsgReg(tx->cob, 3 ) , (tx->id << 5 ) & 0xff );
  }

  if( tx->flags & MSG_RTR ) {
	iCANout(minor,iMsgReg(tx->cob, 6 ) , (tx->length << 4) | 0x00 | (xtd << 2));
  } else {
	iCANout(minor,iMsgReg(tx->cob, 6 ) , (tx->length << 4) | 0x08 | (xtd << 2));
	for ( i = 0; i < tx->length; i++ ) {
      		iCANout(minor,iMsgReg(tx->cob, 7+i ) , tx->data[i] );
        }
  }

  iCANout(minor,iMsgReg(tx->cob, 0 ),  MSGVAL_SET | TXIE_NOP   | RXIE_NOP   | INTPND_RST );
  iCANout(minor,iMsgReg(tx->cob, 1 ),  RMTPND_NOP | TXRQST_SET | CPUUPD_RST | NEWDAT_SET );
  
  if(wait) LDDK_STOP_TIMER();
  DBGout();return i;

}/



func i527_GetMessage(int minor, canmsg_t *rx ) /{

}/


