header "util"
/{ 

 extern volatile int irq2minormap[];
 extern volatile int irq2pidmap[];

 #define MAX_BUFS 4
 #define MAX_BUFSIZE 64

 #define BUF_EMPTY    0
 #define BUF_OK       1
 #define BUF_OVERRUN  2
 #define BUF_UNDERRUN 3

 typedef struct {
	int head;
        int tail;
        int status;
	int active;
        canmsg_t data[MAX_BUFSIZE];
 } msg_fifo_t;

 #define MAX_ID_LENGTH 11
 #define MAX_ID_NUMBER (1<<11)

 typedef struct {
	unsigned    use;
	unsigned    signo[3];
	struct {
		unsigned    enable    : 1;
		unsigned    timestamp : 1;
		unsigned    signal    : 2;
		canmsg_t    *rtr_response;
	} filter[MAX_ID_NUMBER];
 } msg_filter_t;


 extern msg_filter_t Rx_Filter[];

 extern msg_fifo_t Tx_Buf[];
 extern msg_fifo_t Rx_Buf[];

 extern int Can_RequestIrq(int minor, int irq, void (*handler)());
 extern struct wait_queue *CanWait[];

}/

/*======================================================================*/

file "util" /{
 
 volatile int irq2minormap[15];
 volatile int irq2pidmap[15];
 struct wait_queue *CanWait[MAX_BUFS];

 msg_fifo_t   Tx_Buf[MAX_BUFS];
 msg_fifo_t   Rx_Buf[MAX_BUFS];
 msg_filter_t Rx_Filter[MAX_BUFS]; 

 int IRQ_requested[] = {0,0,0,0};

 int Can_RequestIrq(int minor, int irq, void (*handler)()) {
	int err=0;
	DBGin("Can_RequestIrq");
	if( VendOpt[minor] == 's' ){
		err = request_irq( irq, handler, SA_INTERRUPT, "%<driver.name>", NULL );
	} else {
		err = request_irq( irq, handler, SA_SHIRQ, "%<driver.name>", NULL );
	}
	if( !err ){
	  irq2minormap[irq] = minor;
	  irq2pidmap[irq] = current->pid;
	  DBGprint(DBG_BRANCH,("Requested IRQ: %d",irq));
	  IRQ_requested[minor] = 1;
	}
	DBGout();return err;
 }

 int Can_FreeIrq(int minor, int irq ) {
	DBGin("Can_FreeIrq");
        IRQ_requested[minor] = 0;
	free_irq(irq,NULL);
	DBGout();
 }

 int Can_WaitInit(int minor) {
	int i;
	DBGin("Can_WaitInit");
		CanWait[minor] = NULL;
	DBGout();
 }

 int Can_FifoInit(int minor) {
	DBGin("Can_FifoInit");
	   Tx_Buf[minor].head   = Rx_Buf[minor].head = 0;
	   Tx_Buf[minor].tail   = Rx_Buf[minor].tail = 0;
	   Tx_Buf[minor].status = Rx_Buf[minor].status = 0;
	   Tx_Buf[minor].active = Rx_Buf[minor].active = 0;
	DBGout();
 }

 int Can_FilterInit(int minor) {
	int i;
	DBGin("Can_FilterInit");
	   Rx_Filter[minor].use      = 0;
	   Rx_Filter[minor].signo[0] = 0;
	   Rx_Filter[minor].signo[1] = 0;
	   Rx_Filter[minor].signo[2] = 0;

	   for(i=0;i<MAX_ID_NUMBER;i++)	
	      Rx_Filter[minor].filter[i].rtr_response = NULL;

	DBGout();
 }

 int Can_FilterCleanup(int minor)  {
	int i;
	DBGin("Can_FilterCleanup");
	for(i=0;i<MAX_ID_NUMBER;i++) {
		if( Rx_Filter[minor].filter[i].rtr_response != NULL )	
	           kfree( Rx_Filter[minor].filter[i].rtr_response);
		Rx_Filter[minor].filter[i].rtr_response = NULL;
	}
	DBGout();
 }

 int Can_ConfigRTR ( int minor, unsigned message, canmsg_t *Tx ) {
	canmsg_t *tmp;
	DBGin("Can_ConfigRTR");
	if( (tmp = kmalloc ( sizeof(canmsg_t), GFP_ATOMIC )) == NULL ){
		DBGprint(DBG_BRANCH,("memory problem"));
		DBGout(); return -1;
	}
	Rx_Filter[minor].filter[message].rtr_response = tmp;
	memcpy( Rx_Filter[minor].filter[message].rtr_response , Tx, sizeof(canmsg_t));	
	DBGout(); return 1;
 }

 int Can_UnConfigRTR ( int minor, unsigned message ) {
	canmsg_t *tmp;
	DBGin("Can_UnConfigRTR");
	if( Rx_Filter[minor].filter[message].rtr_response != NULL ) {
		kfree(Rx_Filter[minor].filter[message].rtr_response);
		Rx_Filter[minor].filter[message].rtr_response = NULL;
	}	
	DBGout(); return 1;
 }

 int Can_FilterOnOff(int minor, unsigned on) {
	DBGin("Can_FilterOnOff");
	   Rx_Filter[minor].use = (on!=0);
	DBGout();
 }

 int Can_FilterMessage(int minor, unsigned message, unsigned enable) {
	DBGin("Can_FilterMessage");
	   Rx_Filter[minor].filter[message].enable = (enable!=0);
	DBGout();
 }

 int Can_FilterTimestamp(int minor, unsigned message, unsigned stamp) {
	DBGin("Can_FilterTimestamp");
	   Rx_Filter[minor].filter[message].timestamp = (stamp!=0);
	DBGout();
 }

 int Can_FilterSignal(int minor, unsigned id, unsigned signal) {
	DBGin("Can_FilterSignal");
	   if( signal <= 3 )
	   Rx_Filter[minor].filter[id].signal = signal;
	DBGout();
 }

 int Can_FilterSigNo(int minor, unsigned signo, unsigned signal ) {
	DBGin("Can_FilterSigNo");
	   if( signal < 3 )
	   Rx_Filter[minor].signo[signal] = signo;
	DBGout();
 }



}/



