#ifdef HAS_READLINE
#include <readline.h>
#include <history.h>
#endif


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <Can_lib.h>

#define STDDEV "can1"

#define HAS_READLINE



main(int argc,char **argv) {
int fd;
int id;
int terminate=0;
int i,j,got;
canmsg_t rx[80];
canmsg_t tx[80];
struct timeval tv;

#ifndef HAS_READLINE
char command[80];
#else
char *command;
#endif

char *token;
int base=0;

char device[40];
fd_set rfds;
pid_t pid;


printf("\
   Welcome to CANterm (c) 1998 by C.Schroeter (clausi@chemie.fu-berlin.de)\n\
     please type 'help' for command info and 'q' for exit. \n\
   \n\   
");



    if(argc >= 2) {
	sprintf(device, "/dev/%s", argv[1]);
    }
    else {
	sprintf(device, "/dev/%s", STDDEV);
    }
    printf("using CAN device %s\n", device);

if(( fd = open(device, O_RDWR )) < 0 ) {
  fprintf(stderr,"Error opening CAN device %s\n", device);
  exit(1);
}

if( ( pid=fork() ) < 0 ) {
  fprintf(stderr,"\n fork error ");
  exit(1);
} else if (pid==0) {                   /* child */
  if( argc >= 3 ) {
    for(i = 2;i<argc;i++) {
      id = atoi(argv[i]);
      printf("setting filter on id %d\n",id);
      can_Config(fd, CONF_FILTER, 1);
      can_Config(fd, CONF_FENABLE, id );
    }
  }

  /*printf("waiting on %d for msg at 0x%lx:\n",fd, &rx);*/

  tv.tv_sec  = 5; /* wait 5 seconds before process terminates */
  tv.tv_usec = 0;

  while(1) {
    
    FD_ZERO(&rfds);
    FD_SET(fd,&rfds);
    got=0;
    if( select(FD_SETSIZE, &rfds, NULL, NULL, NULL  ) > 0 && FD_ISSET(fd,&rfds) ) {

      got=read(fd, rx , 79 );
      if( got > 0) {

	printf("\n");
     
	for(i=0;i<got;i++) {
	  rx[i].data[rx[i].length] = 0;
	  if( rx[i].flags & MSG_OVR ) printf("OVERRUN: ");

	  printf("Received got=%d: id=%d len=%d flags=0x%x ",
		 got, rx[i].id, rx[i].length, rx[i].flags );
	  if( rx[i].flags & MSG_RTR )
	    printf(" RTR \n\t") ;
	  else
	    printf(" msg='%s' \n\t", rx[i].data );      
	  for(j=0;j<rx[i].length;j++)
	    printf(" 0x%x",rx[i].data[j] );
	  printf("\n");
	  fflush(stdout); }
      } else {
	printf("- Received got=%d: id=%d len=%d msg='%s' \n",
	       got, rx[i].id, rx[i].length, rx[i].data );
	fflush(stdout); }
      /*sleep(1);*/
        

    }
    if (got>0 ) printf ("\nCommand>"); fflush(stdout);
    if(terminate) abort();
  }

} else {                               /* parent */

  while(1) {
#ifndef HAS_READLINE
   fprintf(stderr,"\nCmd>");
   fgets(command,79,stdin);
#else
   command = readline("Command>");
#endif
   if( command[0] == 'q' ) {
            printf("exiting...\n");
            terminate = 1;
	    waitpid(pid, NULL, 0);
            exit(0);
   }
   token=strtok(command," ");
   if( token != NULL )
   if( token[0] == 'q' && !strncmp(token,"quit",4) ) {
            printf("exiting...\n");
            exit(0);
   }
   else if( token[0] == 's' && !strncmp(token,"send",4) ) {
     int id;
     char msg[80];
     int sent=0;

     if((token=strtok(NULL," ") )!= NULL ) {
       sscanf(token,"%d:%s",&id,msg);
       printf("\n ID=%d msg=%s",id,msg); fflush(stdout);
       tx[0].flags=0;
       tx[0].length=strlen(msg);
       tx[0].id=id;
       strcpy(tx[0].data,msg);
       sent = write(fd,&tx[0],1);
       if(sent > 0 ) printf(" OK\n"); fflush(stdout);
     } else {
       printf("usage: send id:string \n");
     }
   }
   else if( token[0] == 'R' && !strncmp(token,"RTR",3) ) {
     int id;
     char msg[80];
     int sent=0;

     if((token=strtok(NULL," ")) != NULL) {
       sscanf(token,"%d",&id);
       printf("\n ID=%d RTR",id,msg); fflush(stdout);
       tx[0].flags= MSG_RTR;
       tx[0].id = id;
       sent = write(fd,&tx[0],1);
       if(sent > 0 ) printf(" OK\n"); fflush(stdout);
     } else {
       printf("usage: RTR id \n");
     }
   }
   else if( token[0] == 'b' && !strncmp(token,"bsend",5) ) {
     int id,j;
     char msg[80];
     char *token;
     int sent=0;

     if((token = strtok(NULL,":")) != NULL) {
       if( token != NULL ) {
	 sscanf( token ,"%d",&id);
	 printf("\n ID=%d ",id);
	 j=0;
	 while( (token = strtok(NULL,",")) != NULL ) {
	   printf("'%c' ",strtol(token,NULL,0) );
	   tx[0].data[j++] = strtol(token,NULL,0);
	 }
	 tx[0].flags=0;
	 tx[0].length=j;
	 tx[0].id=id;
	 sent = write(fd,&tx[0],1);
	 if(sent > 0 ) printf(" OK\n"); fflush(stdout);
       }
     } else {
       printf("usage: bsend id:binary[,binary...] \n");
     }
   }
  fflush(stdout);
#ifdef HAS_READLINE
  free(command);
#endif
  }
}

close(fd);
exit(0);

}












